/*
 * Takes care of the device node located in /dev.
 */

#ifndef UDEV_HANDLER_C
#define UDEV_HANDLER_C

#include <linux/device.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/printk.h>

// Populate memory for the variables that need to be stored

/* The major udev number, used to identify the device file */
static int udev_major_number;
/* The udev device class. */
static struct class* sevenseg_class;
/* The udev device driver */
static struct device* sevenseg_device;
/* Stores the udev's initialization state. */
static bool initialized = false;

// Define prototype functions for file operations
static int sevenseg_open(struct inode *, struct file *);
static int sevenseg_release(struct inode *, struct file *);
static ssize_t sevenseg_read(struct file *, char *, size_t, loff_t *);
static ssize_t sevenseg_write(struct file *, const char *, size_t, loff_t *);

// Store these prototype methods in a structire
static struct file_operations sevenseg_fops = {
    .open    = sevenseg_open,
    .read    = sevenseg_read,
    .write   = sevenseg_write,
    .release = sevenseg_release
};

/* Initializes the udev node */
int sevenseg_udev_init(void)
{
    printk(KERN_DEBUG "sevenseg: Initializing udev interface.\n");

    udev_major_number = register_chrdev(0, DEVICE_NAME, &sevenseg_fops);
    if (udev_major_number < 0)
    {
        printk(KERN_ERR "sevenseg: Error: Failed to initialize the udev node! Exiting.\n");
        return udev_major_number;
    }

    printk(KERN_DEBUG "sevenseg: udev interface initialized successfully with major number %i.\n", udev_major_number);

    // Register the device class
    sevenseg_class = class_create(THIS_MODULE, CLASS_NAME);
    if (IS_ERR(sevenseg_class))
    {
        unregister_chrdev(udev_major_number, DEVICE_NAME);
        printk(KERN_ERR "sevenseg: Error: Failed to register device class! Exiting.\n");
        return PTR_ERR(sevenseg_class);
    }
    printk(KERN_DEBUG "sevenseg: Device class registered successfully.\n");

    // Register the device driver
    sevenseg_device = device_create(sevenseg_class, NULL, MKDEV(udev_major_number, 0), NULL, DEVICE_NAME);
    if (IS_ERR(sevenseg_device))
    {
        class_destroy(sevenseg_class);
        unregister_chrdev(udev_major_number, DEVICE_NAME);
        printk(KERN_ERR "sevenseg: Error: Failed to create the device! Did /dev/sevenseg already exist? Exiting.\n");
        return PTR_ERR(sevenseg_device);
    }

    printk(KERN_INFO "sevenseg: Success! Listening for numbers to display on /dev/sevenseg.\n");
    initialized = true;
    return 0;
}

/* Closes the udev node. */
void sevenseg_udev_exit(void)
{
    printk(KERN_DEBUG "sevenseg: Exiting udev interface.\n");
    device_destroy(sevenseg_class, MKDEV(udev_major_number, 0));
    class_unregister(sevenseg_class);
    class_destroy(sevenseg_class);
    unregister_chrdev(udev_major_number, DEVICE_NAME);
    printk(KERN_DEBUG "sevenseg: udev interface closed.\n");
    initialized = false;
}

/*
 * Executed when the device file has been opened.
 * 
 * @param inodep A pointer to the inode object
 * @param filep A pointer to the file object
 */
int sevenseg_open(struct inode *inodep, struct file *filep)
{
    if (!initialized)
    {
        printk(KERN_ALERT "sevenseg: Tried to open /dev/sevenseg without having it initialized first!\n");
        return -1;
    }
    printk(KERN_DEBUG "sevenseg: Device node has been opened.\n");
    return 0;
}

/*
 * Executed when data has been read from the device file.
 * 
 * @param filep A pointer to the file object
 * @param buffer The character buffer that has been read
 * @param length The buffer's length
 * @param offset The buffer offset (if required)
 */
static ssize_t sevenseg_read(struct file *filep, char *buffer, size_t length, loff_t *offset)
{
    printk(KERN_DEBUG "sevenseg: Read character \"%c\"\n", buffer[0]);
    return length;
}

/*
 * Executed when data has been written to the device file.
 * 
 * @param filep A pointer to the file object
 * @param buffer The character buffer that has been written
 * @param length The buffer's length
 * @param offset The buffer offset (if required)
 */
static ssize_t sevenseg_write(struct file *filep, const char *buffer, size_t length, loff_t *offset)
{
    printk(KERN_DEBUG "sevenseg: Written character \"%c\" to /dev/sevenseg.\n", buffer[0]);
    sevenseg_gpio_update(char_to_int(buffer[0]));
    return length;
}

/*
 * Executed when the device file has been closed.
 * 
 * @param inodep A pointer to the inode object
 * @param filep A pointer to the file object
 */
static int sevenseg_release(struct inode *inodep, struct file *filep)
{
    if (!initialized)
    {
        printk(KERN_ALERT "sevenseg: Tried to close /dev/sevenseg without having it initialized first!\n");
        return -1;
    }
    printk(KERN_DEBUG "sevenseg: Device node has been closed.\n");
    return 0;
}

#endif